#ifndef __GPUDB__
#define __GPUDB__

#include <shrUtils.h>
#include <cutil_inline.h>
#include <gpuDB.h>
#include <gpuDB_kernels.cuh>

#define BLOCK_SIZE 512		// Size of the thread blocks
#define STRIDE_SIZE 128		// Approximate number of records each thread is going to handle
#define HASH_ENTRIES 1024

void computeParallelHashTable(unsigned int size, uint2 *data, Table &table, unsigned int rangeG)
{
	unsigned int deviceID;
	deviceID = cutGetMaxGflopsDeviceId();
	cudaSetDevice(deviceID);
	shrLog("Using device: %d\n", deviceID);

	float elapsedTime;
	int *d_hashLocks, *d_freeLock, *d_firstFree;
	uint2 *d_idata;
	cudaEvent_t start, stop;
	Table gpuTable;

	unsigned int mem_size = sizeof(uint2) * size;
	shrLog("Allocating %d MB for input data\n", (mem_size + 1048576 - 1)/1048576);

	// Initialize events
	cudaEventCreate(&start);
	cudaEventCreate(&stop);

	cudaEventRecord(start, 0);
	cutilSafeCall(cudaMalloc((void**)&d_idata, mem_size));
	cutilSafeCall(cudaMalloc((void**)&d_hashLocks, HASH_ENTRIES * sizeof(int)));
	cutilSafeCall(cudaMalloc((void**)&d_freeLock, sizeof(int)));
	cutilSafeCall(cudaMalloc((void**)&d_firstFree, sizeof(int)));
	
	initialize_GPUTable(gpuTable, (unsigned int)HASH_ENTRIES, rangeG);

	cutilSafeCall(cudaMemcpy(d_idata, data, mem_size, cudaMemcpyHostToDevice));
	cutilSafeCall(cudaMemset(d_hashLocks, 0, HASH_ENTRIES * sizeof(int)));
	cutilSafeCall(cudaMemset(d_freeLock, 0, sizeof(int)));
	cutilSafeCall(cudaMemset(d_firstFree, 0, sizeof(int)));

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&elapsedTime, start, stop);

	shrLog("Host to device: %.3f ms\n", elapsedTime);

	dim3 threads(BLOCK_SIZE, 1);
	//dim3 grid((size + BLOCK_SIZE - 1)/(BLOCK_SIZE * STRIDE_SIZE), 1);
	dim3 grid(2, 1);

	cudaFuncSetCacheConfig(computeAggregation, cudaFuncCachePreferShared);

	cudaEventRecord(start, 0);
	computeAggregation<<<grid, threads>>>(size, d_idata, gpuTable, d_hashLocks, d_freeLock, d_firstFree);

	//cudaThreadExit();

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&elapsedTime, start, stop);

	cutilCheckMsg("Kernel execution failed");

	shrLog("Parallel runtime: %.3f ms\n", elapsedTime);

	float throughput = (float)size / (elapsedTime / 1000);
	shrLog("Throughput: %.3f tuples/s\n", throughput);

	cudaEventRecord(start, 0);
	copy_table_to_host(gpuTable, table, rangeG);
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&elapsedTime, start, stop);

	shrLog("Device to host: %.3f ms\n", elapsedTime);

	free_GPUTable(gpuTable);
	cutilSafeCall(cudaFree(d_firstFree));
	cutilSafeCall(cudaFree(d_freeLock));
	cutilSafeCall(cudaFree(d_hashLocks));
	cutilSafeCall(cudaFree(d_idata));
}

void computeSerial(unsigned int size, uint2 *data, unsigned int *count, unsigned int *sum, unsigned int *sumSquared)
{
	unsigned int i, g, v;
	for (i = 0; i < size; i++) {
		g = data[i].x;
		v = data[i].y;
		count[g]++;
		sum[g] += v;
		sumSquared[g] += v * v;
	}
}

void copy_table_to_host(const Table &table, Table &hostTable, int elements) {
	hostTable.count = table.count;
	hostTable.entries = (int *)calloc(table.count, sizeof(int));
	hostTable.pool = (Entry*)malloc(elements * sizeof(Entry));

	cutilSafeCall(cudaMemcpy(hostTable.entries, table.entries, table.count * sizeof(int), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy(hostTable.pool, table.pool, elements * sizeof(Entry), cudaMemcpyDeviceToHost));
}

void free_GPUTable(Table &table) {
	cutilSafeCall(cudaFree(table.pool));
	cutilSafeCall(cudaFree(table.entries));
}

void free_table(Table &table)
{
	free(table.entries);
	free(table.pool);
}

void initialize_GPUTable(Table &table, unsigned int entries, unsigned int elements)
{
	table.count = entries;

	cutilSafeCall(cudaMalloc((void**)&table.entries, entries * sizeof(int)));
	cutilSafeCall(cudaMemset(table.entries, -1, entries * sizeof(int)));
	cutilSafeCall(cudaMalloc((void**)&table.pool, elements * sizeof(Entry)));
}

int main(int argc, char** argv) 
{
	float elapsedTime;
	Table parallelTable;
	clock_t start, stop;

	// Start logs
	shrSetLogFileName ("gpuDB.txt");
	shrLog("%s Starting...\n\n", argv[0]);

	shrLog("Parsing command line arguments...\n\n");
	if (!parseCommandLineArguments(argc, (const char**)argv)) {
		shrEXIT(argc, (const char**)argv);
		return 1;
	}

	shrLog("Done parsing...\n\n");

	data = (uint2 *)malloc(size * sizeof(uint2));
	if (data == NULL) {
		shrLog("Error allocating memory for data...\n\n");
		shrEXIT(argc, (const char**)argv);
		return 1;
	}

	serialCount = (unsigned int *)calloc(rangeG, sizeof(unsigned int));
	if (serialCount == NULL) {
		shrLog("Error allocating memory for serial count...\n\n");
		free(data);
		shrEXIT(argc, (const char**)argv);
		return 1;
	}

	serialSum  = (unsigned int *)calloc(rangeG, sizeof(unsigned int));
	if (serialSum == NULL) {
		shrLog("Error allocating memory for serial sum...\n\n");
		free(serialCount);
		free(data);
		shrEXIT(argc, (const char**)argv);
		return 1;
	}

	serialSumSquared = (unsigned int *)calloc(rangeG, sizeof(unsigned int));
	if (serialSumSquared == NULL) {
		shrLog("Error allocating memory for serial sum squared...\n\n");
		free(serialSum);
		free(serialCount);
		free(data);
		shrEXIT(argc, (const char**)argv);
		return 1;
	}

	populate(data);

	if (!test) {
		start = clock();
		
		// Compute and time serial naive aggregation
		start = clock();
		computeSerial(size, data, serialCount, serialSum, serialSumSquared);
		stop = clock();
		elapsedTime = (float)(stop - start) / (float) CLOCKS_PER_SEC * 1000.0f;
		shrLog("Time to compuete serial results: %.3f ms\n", elapsedTime);

		// Compute and time parallel GPU aggregation
		start = clock();
		computeParallelHashTable(size, data, parallelTable, rangeG);
		stop = clock();
		elapsedTime = (float)(stop - start) / (float) CLOCKS_PER_SEC * 1000.0f;
		shrLog("Time to compute parallel hash: %.3f ms\n", elapsedTime);

		verify_table(parallelTable, "Parallel");

		free_table(parallelTable);
		free(serialSumSquared);
		free(serialSum);
		free(serialCount);
	} else {
		testGPU(size, data);
	}

	free(data);
	
	shrEXIT(argc, (const char**)argv);
	return 0;
}

bool
parseCommandLineArguments(const int argc, const char** argv)
{
	int errors = 0;

	if (shrCheckCmdLineFlag(argc, argv, "seed"))
		if (!shrGetCmdLineArgumentu(argc, argv, "seed", &seed))
			errors++;

	if (shrCheckCmdLineFlag(argc, argv, "rangeG"))
		if (!shrGetCmdLineArgumentu(argc, argv, "rangeG", &rangeG))
			errors++;

		if (shrCheckCmdLineFlag(argc, argv, "rangeV"))
		if (!shrGetCmdLineArgumentu(argc, argv, "rangeV", &rangeV))
			errors++;

	if (shrCheckCmdLineFlag(argc, argv, "size"))
		if (!shrGetCmdLineArgumentu(argc, argv, "size", &size))
			errors++;

	if (shrCheckCmdLineFlag(argc, argv, "input"))
		if (!shrGetCmdLineArgumentstr(argc, argv, "input", &inputFile))
			errors++;

	if (shrCheckCmdLineFlag(argc, argv, "output"))
		if (!shrGetCmdLineArgumentstr(argc, argv, "outputFile", &outputFile))
			errors++;

	if (shrCheckCmdLineFlag(argc, argv, "test"))
		if (!shrGetCmdLineArgumenti(argc, argv, "test", (int *)&test))
			errors++;

	if (shrCheckCmdLineFlag(argc, argv, "verbose"))
		if (!shrGetCmdLineArgumenti(argc, argv, "verbose", (int *)&verbose))
			errors++;

	if (errors > 0) {
		shrLog("%d errors(s) in parsing command arguments\n\n", errors);
		printUsage();
		return false;
	}

	return true;
}

void
populate(uint2 *data)
{
	unsigned int i;
	srand(seed);
	for (i = 0; i < size; i++) {
		data[i].x = rand() % rangeG;
		data[i].y = rand() % rangeV;
	}
}

void
printUsage()
{
#ifdef _WIN32
	shrLog("Usage: gpuDB.exe [optional: --argument=value]\n");
#else
	shrLog("Usage: gpuDB [optional: --argument=value]\n");
#endif
	shrLog("\tseed     :  Seed to use for random generator (any positive integer, default is current time in seconds)\n");
	shrLog("\trangeG   :  Range of values to group by (any positive integer, default is 100)\n");
	shrLog("\trangeV   :  Range of values in each ercord (any positive integer, default is 100)\n");
	shrLog("\tsize     :  Number of records (any positive integer, default is 1000)\n");
	shrLog("\tinput    :  File containing input to utilize for computation\n");
	shrLog("\toutput   :  File containing serial output to compare against (if present, serial computation is skipped)\n");
	shrLog("\ttest     :  Test the GPU by running a simple test kernel\n\n");
	shrLog("\tverbose  :  0 = standard runtime, 1 = verbose runtime\n\n");
}

void testGPU(unsigned int size, uint2 *data)
{
	uint2 *d_idata, *d_odata, *odata;
	unsigned int mem_size = sizeof(uint2) * size;

	odata = (uint2 *)malloc(mem_size);
	cutilSafeCall(cudaMalloc((void**)&d_idata, mem_size));
	cutilSafeCall(cudaMalloc((void**)&d_odata, mem_size));

	cutilSafeCall(cudaMemcpy(d_idata, data, mem_size, cudaMemcpyHostToDevice));
	
	dim3 threads(STRIDE_SIZE, 1);
	dim3 grid((size + STRIDE_SIZE - 1)/STRIDE_SIZE, 1);

	testKernel<<<grid, threads>>>(size, d_idata, d_odata);

	cutilCheckMsg("Kernel execution failed");

	cutilSafeCall(cudaMemcpy(odata, d_odata, mem_size, cudaMemcpyDeviceToHost));

	for (unsigned int i = 0; i < size; i++)
		if (odata[i].x != data[size - 1 - i].x && odata[i].y != data[size - 1 - i].y)
			shrLog("Error in test on element %d\n", i);

	shrLog("Kernel test complete!\n");

	free(odata);
	cutilSafeCall(cudaFree(d_odata));
	cutilSafeCall(cudaFree(d_idata));
}

void verify_table(const Table &table, const char *type)
{
	unsigned int key;
	unsigned int count = 0;
	for (unsigned int i = 0; i < table.count; i++) {
		int EntryIndex = table.entries[i];
		while (EntryIndex != -1) {
			Entry current = table.pool[EntryIndex];
			key = current.data.x;
			count += current.data.y;
			
			if (hash(key, table.count) != i)
				shrLog("%s: %d hashed to %d, but was located at %d\n", type, key, hash(key, table.count), i);
			if (current.data.y != serialCount[key])
				shrLog("%s: %d hash table entry count %d does not match serial count %d\n", type, key, current.data.y, serialCount[key]);
			if (current.data.z != serialSum[key])
				shrLog("%s: %d hash table entry sum %d does not match serial sum %d\n", type, key, current.data.z, serialSum[key]);
			if (current.data.w != serialSumSquared[key])
				shrLog("%s: %d hash table entry sum squared %d does not match serial sum squared %d\n", type, key, current.data.w, serialSumSquared[key]);
			
			EntryIndex = current.next;
		}
	}
	
	if (count != size)
		shrLog("%s: %d elements found in hash table. Should be %d\n", type, count, size);
	else
		shrLog("%s: All %d elements found in hash table. \n", type, count);
}

#endif
