#ifndef __GPUDB_KERNELS__
#define __GPUDB_KERNELS__

#include <gpuDB_kernels.cuh>

#define HASH_ENTRIES 1024
#define HASH_SIZE 512

__device__ void lock(int *mutex) {
	while(atomicCAS(mutex, 0, 1) != 0);
}

__device__ void unlock(int *mutex) {
	atomicExch(mutex, 0);
}

__device__ void add_to_global_hash_table(unsigned int key, unsigned int count, unsigned int sum, unsigned int sumSquared, Table table, int *globalHashLocks, int *globalFreeLock, int *globalFirstFree)
{
	// Find entry if it exists
	unsigned int hashValue = hash(key, table.count);
	
	lock(&globalHashLocks[hashValue]);
	__threadfence();

	int bucketHead = table.entries[hashValue];
	int currentLocation = bucketHead;
	
	bool found = false;
	Entry currentEntry;

	while (currentLocation != -1 && !found) {
		currentEntry = table.pool[currentLocation];
		if (currentEntry.data.x == key) {
			found = true;
		} else {
			currentLocation = currentEntry.next;
		}
	}
	
	if (currentLocation == -1) {
		// If entry does not exist, create entry
		lock(globalFreeLock);
		int newLocation = (*globalFirstFree)++;
		unlock(globalFreeLock);
		__threadfence();

		Entry newEntry;
		newEntry.data.x = key;
		newEntry.data.y = count;
		newEntry.data.z = sum;
		newEntry.data.w = sumSquared;
		newEntry.next = bucketHead;

		// Add entry to table
		table.pool[newLocation] = newEntry;
		__threadfence();
		table.entries[hashValue] = newLocation;
		__threadfence();
	} else {
		currentEntry.data.y += count;
		currentEntry.data.z += sum;
		currentEntry.data.w += sumSquared;
		table.pool[currentLocation] = currentEntry;
		__threadfence();
	}

	unlock(&globalHashLocks[hashValue]);
	__threadfence();
}

__device__ void add_to_shared_hash_table(unsigned int key, unsigned int value, int *sharedHashLocks, int *sharedFreeLock, int *entries, Entry *pool, int *sharedFirstFree)
{
	// Find entry in shared hash if it exists
	size_t sharedHashValue = hash(key, HASH_ENTRIES);
	
	lock(&sharedHashLocks[sharedHashValue]);
	__threadfence();

	int bucketHead = entries[sharedHashValue];
	int currentLocation = bucketHead;
	
	bool found = false;
	Entry currentEntry;

	while (currentLocation != -1 && !found) {
		currentEntry = pool[currentLocation];
		if (currentEntry.data.x == key) {
			found = true;
		} else {
			currentLocation = currentEntry.next;
		}
	}

	if (currentLocation == -1) {
		// If entry does not exist, create entry
		lock(sharedFreeLock);
		int newLocation = (*sharedFirstFree)++;
		unlock(sharedFreeLock);
		__threadfence();

		Entry newEntry;
		newEntry.data.x = key;
		newEntry.data.y = 1;
		newEntry.data.z = value;
		newEntry.data.w = (value * value);
		newEntry.next = bucketHead;

		// Add entry to table
		pool[newLocation] = newEntry;
		__threadfence();
		entries[sharedHashValue] = newLocation;
		__threadfence();
	} else {
		currentEntry.data.y++;
		currentEntry.data.z += value;
		currentEntry.data.w += (value * value);
		pool[currentLocation] = currentEntry;
		__threadfence();
	}

	unlock(&sharedHashLocks[sharedHashValue]);
	__threadfence();
}


__global__ void computeAggregation(unsigned int size, uint2 *idata, Table table, int *globalHashLocks, int *globalFreeLock, int *globalFirstFree)
{
	__shared__ int sharedHashLocks[HASH_ENTRIES];
	__shared__ int sharedFreeLock;
	__shared__ int entries[HASH_ENTRIES];
	__shared__ Entry pool[HASH_SIZE];
	__shared__ int sharedFirstFree;

	uint2 tuple;
	Entry sharedEntry;
	unsigned int index, warpThread;
	unsigned int threadID = (blockIdx.x * blockDim.x) + threadIdx.x;
	unsigned int stride = (gridDim.x * blockDim.x);

	if (threadIdx.x == 0) {
		sharedFreeLock = 0;
		sharedFirstFree = 0;
	}
	__syncthreads();

	for (index = threadIdx.x; index < HASH_ENTRIES; index += blockDim.x) {
			sharedHashLocks[index] = 0;
			entries[index] = -1;
	}
	__syncthreads();

	for (index = threadID; index < size; index += stride) {
		tuple = idata[index];
		for (warpThread = 0; warpThread < warpSize; warpThread++) {
			if ((threadIdx.x % warpSize) == warpThread) {
				add_to_shared_hash_table(tuple.x, tuple.y, sharedHashLocks, &sharedFreeLock, entries, pool, &sharedFirstFree);
			}
		}
		__syncthreads();
	}
	__syncthreads();

	for (index = threadIdx.x; index < HASH_SIZE; index += blockDim.x) {
		sharedEntry = pool[index];
		for (warpThread = 0; warpThread < warpSize; warpThread++) {
			if ((threadIdx.x % warpSize) == warpThread) {
				add_to_global_hash_table(sharedEntry.data.x, sharedEntry.data.y, sharedEntry.data.z, sharedEntry.data.w, table, globalHashLocks, globalFreeLock, globalFirstFree);
			}
		}
		__syncthreads();
	}
}

__device__ __host__ unsigned int hash(unsigned int key, unsigned int count)
{
	return key % count;
}

__global__ void testKernel(unsigned int size, uint2 *idata, uint2 *odata)
{
	unsigned int threadID = (blockIdx.x * blockDim.x) + threadIdx.x;

	if (threadID < size) {
		odata[size - 1 - threadID] = idata[threadID];
	}
}

#endif
